/*
** Module   : PostgreSQLSequenceHandler.java
** Abstract : Static class that handles sequence functions and statements 
**            from Progress / OpenEdge 10.x.
**
** Copyright (c) 2012-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 OM  20121116 Created initial version with CURRVAL, NEXTVAL and SETVAL support only.
** 002 SVL 20120331 Upgraded to Hibernate 4.
** 003 OM  20120712 Fixed extraction data from ResultSets. Added error handling.
*/

package com.goldencode.p2j.persist.sequence;

import java.util.*;
import org.hibernate.*;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.persist.dialect.*;
import com.goldencode.p2j.util.*;

/**
 * PostgreSQL implementation of sequences using a native SQL sequence
 * primitives queries.
 */
public class PostgreSQLSequenceHandler
extends SequenceHandler
{

   /**
    * The list of sequences grouped by databases.
    * It uses lazy initialization.
    */
   private HashMap<String, LinkedList<String>> sequences = null;
   
   /**
    * The only constructor takes the name of the database as argument. 
    * It is passed on to super class.
    * 
    * @param   ldbName
    *          The database logical name on which this handler will work with. 
    */
   public PostgreSQLSequenceHandler(String ldbName)
   {
      super(ldbName);
      // TODO Auto-generated constructor stub
   }
   
   /**
    * Queries the current value of a sequence.
    * <p>
    * This method should be called only in a safe environment, where the database-level lock for
    * access to sequences has already been acquired.
    * 
    * @param   seqName
    *          The name of the sequence.
    * @param   ldbName
    *          The database to which the sequence belongs to. It is assumed it is not null and
    *          valid.
    * 
    * @return  the current value of the sequence
    */
   @Override
   protected int64 safeGetCurrentValue(String seqName, String ldbName)
   {
      Sequence seq = SequenceManager.getSequence(ldbName, seqName);
      if (seq == null)
      {
         ErrorManager.recordOrThrowError(2913,
                 "Unknown sequence expression " + seqName.toUpperCase());
         return new int64();
      }
      
      // this will also create the instance if needed:
      Persistence persistence = ConnectionManager.getPersistence(ldbName);
      P2JDialect dialect = (P2JDialect) persistence.getDialect();
      String strQuery = dialect.getSequenceCurrValString(seqName);

      try
      {
         return getResultValue(persistence.executeSQLQuery(strQuery));
      }
      catch (PersistenceException pe)
      {
         pe.printStackTrace();
      }

      // if here then some error occurred, return unknown value
      return new int64();
   }

   /**
    * Computes and returns the next value of a sequence.
    * <p>
    * This method should be called only in a safe environment, where the database-level lock for
    * access to sequences has already been acquired.
    * 
    * @param   seqName
    *          The name of the sequence.
    * @param   ldbName
    *          The database to which the sequence belongs to. It is assumed it is not null and
    *          valid.
    * 
    * @return  the next value of the sequence or unknown value (?) if sequence/database not found
    *          or the sequence is not cycling and has passed the maximum/minimum value.
    */
   @Override
   protected int64 safeGetNextValue(String seqName, String ldbName)
   {
      Sequence seq = SequenceManager.getSequence(ldbName, seqName);
      if (seq == null)
      {
         ErrorManager.recordOrThrowError(2913,
                 "Unknown sequence expression " + seqName.toUpperCase());
         return new int64();
      }
      
      // PSQL server is different than Progress when cycling:
      //    it will make the counter jump from MAX_VALUE to MIN_VALUE (and the reverse) 
      //    instead, 4GL will jump back to INITIAL_VALUE
      // in PSQL, START clause is only used when declaring/initializing the sequence

      // When INCREMENT value is negative, 4GL will automatically enforce the MAX_VALUE to 
      // be equal to INITIAL_VALUE, this is directly supported by PSQL
      // When Increment is positive, MIN_VALUE is undefined so you can programatically set 
      // the value of the sequence down to Long.MIN_VALUE. There are 3 options:
      //    - define MIN_VALUE of the sequence to be Long.MIN_VALUE. The problem is when 
      //          progress reaches the upper limit it will jump back to Long.MIN_VALUE instead of
      //          INIT_VALUE.
      //    - define MIN_VALUE equal to INIT_VALUE. The problem is that you cannot assign small
      //          values like Long.MIN_VALUE to the sequence (PSQL won't allow to break the MIN)
      //    - handle upper cycling in P2J: let the  sequences's MIN_VALUE = Long.MIN_VALUE.
      //          Detect when UPPER cycling occurs and programatically jump to stored INITAl_VALUE
      
      long crtVal = safeGetCurrentValue(seqName, ldbName).longValue();
      long inc = seq.getIncrement();
      if (seq.isCycle() && inc > 0)
      {
         if (crtVal + inc > seq.getMax() ||     // check crossing the upper bound
             crtVal + inc < crtVal)             // check 64-bit overflow
         {
            long ret = seq.getInitial(); 
            safeSetValue(seqName, ldbName, ret);
            return new int64(ret);
         }
      }
      
      Persistence persistence = ConnectionManager.getPersistence(ldbName);
      P2JDialect dialect = (P2JDialect) persistence.getDialect();
      String strQuery = dialect.getSequenceNextValString(seqName);

      try
      {
         return getResultValue(persistence.executeSQLQuery(strQuery));
      }
      catch (PersistenceException pe)
      {
         boolean outOfBoundsException = false;
         if (pe.getCause() != null)
         {
            String errMessage = pe.getCause().getMessage();
            if (errMessage.contains("reached maximum value") ||
                errMessage.contains("reached minimum value"))
            {
               outOfBoundsException = true;
            }
         }
         if (!outOfBoundsException)
         {
            // some other nasty thing has happen, report it:
            pe.printStackTrace();
         }
      }

      // if here then some error occurred, return unknown value
      return new int64();
   }

   /**
    * Initialize (reset) the current value of a sequence.
    * <p>
    * This method should be called only in a safe environment, where the database-level lock for
    * access to sequences has already been acquired.
    *
    * @param   seqName
    *          The name of the sequence.
    * @param   ldbName
    *          The database to which the sequence belongs to. 
    *          It is assumed it is not null and valid.
    * @param   newVal
    *          The value the sequence will be reseted to.
    */
   @Override
   protected void safeSetValue(String seqName, String ldbName, long newVal)
   {
      Sequence seq = SequenceManager.getSequence(ldbName, seqName);
      if (seq == null)
      {
         ErrorManager.recordOrThrowError(3133,
                 "SYSTEM ERROR: Sequence " + seqName.toUpperCase() + " not found");
         return;
      }
      if (seq.getMin() > newVal || newVal > seq.getMax())
      {
         ErrorManager.recordOrThrowError(3132,
                 "Cannot set sequence " + seq.getLegacyName().toUpperCase() + 
                 " beyond its max/min values");
         return;
      }
      
      Persistence persistence = ConnectionManager.getPersistence(ldbName);
      P2JDialect dialect = (P2JDialect) persistence.getDialect();
      String strQuery = dialect.getSequenceSetValString(seqName, newVal);

      try
      {
         persistence.executeSQLQuery(strQuery);
      }
      catch (PersistenceException pe)
      {
         pe.printStackTrace();
      }
   }

   /**
    * Test whether the database contains a sequence. This is useful to identify the correct
    * database/sequences when a 4GL static function / statement is called without the database
    * parameter.
    * <p>
    * This method should be called only in a safe environment, where the database-level lock for
    * access to sequences has already been acquired.
    * 
    * @param   seqName
    *          The name of the sequence.
    * @param   ldbName
    *          The database to which the sequence belongs to. If is null then the first connected
    *          database is assumed.
    * 
    * @return  true if the sequence is defined in the database
    */
   @Override
   protected boolean safeContains(String seqName, String ldbName)
   {
      // lazy initialization
      if (sequences == null)
      {
         sequences = new HashMap<String, LinkedList<String>>();
      }
         
      // make sure the structure is initialized for this database:
      LinkedList<String> crtSeqList = null;
      if (!sequences.containsKey(ldbName))
      {
         // create the list 
         crtSeqList = new LinkedList<String>();
         sequences.put(ldbName, crtSeqList);

         Persistence persistence = ConnectionManager.getPersistence(ldbName);
         try
         {
            ScrollableResults rs = persistence.executeSQLQuery(
                        "SELECT relname " +
                        "FROM \"pg_class\" " +
                        "WHERE relkind='S';");
            while (rs.next())
            {
               crtSeqList.add((String) rs.get(0));
            }
         }
         catch (PersistenceException pe)
         {
            pe.printStackTrace();
            return false;
         }
      }
      else
      {
         crtSeqList = sequences.get(ldbName);
      }

      // now check whether the sequence is in database
      return crtSeqList.contains(seqName);
   }
}
